from __future__ import annotations

import argparse
import importlib.metadata as impm
import logging
import sys
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import Any, TypeVar

import attrs
import more_itertools as itx
from typing_extensions import overload, override

from snakebids import bidsapp
from snakebids.exceptions import ConfigError, RunError
from snakebids.io.config import write_config
from snakebids.plugins.bidsargs import BidsArgs
from snakebids.plugins.cli_config import CliConfig
from snakebids.plugins.component_edit import ComponentEdit
from snakebids.plugins.pybidsdb import Pybidsdb
from snakebids.snakemake_compat import load_configfile
from snakebids.snakemake_compat import main as snakemake_main
from snakebids.utils.output import (
    prepare_bidsapp_output,
    write_output_mode,
)
from snakebids.utils.utils import to_resolved_path

__all__ = ["SnakemakeBidsApp"]

logger = logging.getLogger(__name__)


SNAKEFILE_CHOICES = [
    "Snakefile",
    "snakefile",
    "workflow/Snakefile",
    "workflow/snakefile",
]


CONFIGFILE_CHOICES = [
    "config/snakebids.yml",
    "config/snakebids.yaml",
    "config/snakebids.json",
    "snakebids.yml",
    "snakebids.yaml",
    "snakebids.json",
    "config.yml",
    "config.yaml",
    "config.json",
    "config/config.json",
    "config/config.yml",
    "config/config.yaml",
]


class SnakemakeHelpAction(argparse.Action):
    """Class for printing snakemake usage in argparse."""

    @override
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ):
        snakemake_main(["-h"])  # type: ignore


def _get_file_paths(
    choices: list[str], file_name: str
) -> Callable[[SnakemakeBidsApp], Path]:
    def wrapper(self: SnakemakeBidsApp):
        for path in choices:
            if (self.snakemake_dir / path).exists():
                return Path(self.snakemake_dir, path)

        msg = f"Error: no {file_name} file found, tried {', '.join(choices)}."
        raise ConfigError(msg)

    return wrapper


_T = TypeVar("_T")


@overload
def _resolve_path(path_candidate: Sequence[Any]) -> list[Any]: ...


@overload
def _resolve_path(path_candidate: _T) -> _T: ...


def _resolve_path(path_candidate: Any) -> Any:
    """Resolve paths or list of paths, or return argument unchanged.

    Parameters
    ----------
    command : list, os.Pathlike, object
        command to run

    Returns
    -------
        If os.Pathlike or list  of os.Pathlike, the same paths resolved.
        Otherwise, the argument unchanged.
    """
    if isinstance(path_candidate, Sequence) and not isinstance(path_candidate, str):
        return [
            _resolve_path(p)  # type: ignore[reportUnknownArgumentType]
            for p in path_candidate  # type: ignore[reportUnknownVariableType]
        ]

    if isinstance(path_candidate, Path):
        return Path(path_candidate).resolve()

    return path_candidate


@attrs.define
class SnakemakeBidsApp:
    """Snakebids app with config and arguments.

    .. currentmodule:: snakebids.plugins

    Loads the :class:`~bidsargs.BidsArgs`, :class:`~cli_config.CliConfig`,
    :class:`~pybidsdb.Pybidsdb`, and :class:`~component_edit.ComponentEdit` plugins
    as dependencies.

    Parameters
    ----------
    snakemake_dir : str | Pathlike[str]
        Root directory of the snakebids app, containing the config file and workflow
        files.
    configfile_path
        Path of config file. By default, looks for ``config.yaml``, ``config.json``,
        ``snakebids.yaml``, or ``snakebids.json`` either within the ``snakemake_dir`` or
        within ``snakemake_dir/config``
    configfile_outpath
        Path of output configfile that will be consumed by the snakemake workflow. This
        is only necessary if the input configfile is not within the snakemake
        directory. If given, it should be the same as the path specified under
        ``configfile: ...` within the ``Snakefile``
    snakefile_path
        Absolute path to the input Snakefile. By default, looks for ``Snakefile`` within
        the ``snakemake_dir`` or ``snakemake_dir/workflow``
    """

    snakemake_dir: Path = attrs.field(converter=to_resolved_path)
    configfile_path: Path | None = attrs.field(
        default=attrs.Factory(
            _get_file_paths(CONFIGFILE_CHOICES, "config"), takes_self=True
        ),
        kw_only=True,
    )
    configfile_outpath: Path | None = attrs.field(default=None, kw_only=True)
    snakefile_path: Path = attrs.field(
        default=attrs.Factory(
            _get_file_paths(SNAKEFILE_CHOICES, "Snakefile"), takes_self=True
        ),
        kw_only=True,
    )

    # use lambda so that it can be patched by fakefs in tests
    cwd: Path = attrs.field(factory=lambda: Path(), init=False)
    """The current working directory for running snakemake.

    This is where the ``.snakemake`` metadata folder will be placed.
    """

    force_output: bool = attrs.field(default=False, init=False)
    """Allow specifying outputs in unrecognized, non-empty directories.

    After the first run, a ``.snakebids`` file will be created in the directory to mark
    the directory as safe for future runs.
    """

    DEPENDENCIES = (
        Pybidsdb(),
        ComponentEdit(),
        CliConfig(),
        BidsArgs(),
    )

    @classmethod
    def create_empty(cls):
        """Create empty instance of plugin."""
        return cls(
            snakemake_dir=Path(),
            configfile_path=None,
            snakefile_path=Path("Snakefile"),
        )

    @bidsapp.hookimpl
    def initialize_config(self, config: dict[str, Any]):
        """Read config file and load into config dict."""
        if self.configfile_path is not None:
            config.update(
                load_configfile(str(self.snakemake_dir / self.configfile_path))
            )

    @bidsapp.hookimpl
    def add_cli_arguments(self, parser: argparse.ArgumentParser):
        """Add snakemake help and force_output arguments."""
        parser.add_argument(
            "--help-snakemake",
            "--help_snakemake",
            nargs=0,
            action=SnakemakeHelpAction,
            help=(
                "Options to Snakemake can also be passed directly at the "
                "command-line, use this to print Snakemake usage"
            ),
        )
        parser.add_argument(
            "--force-output",
            "--force_output",
            action="store_true",
            help="Force output in a new directory that already has contents",
        )

    @bidsapp.hookimpl
    def handle_unknown_args(self, args: list[str], config: dict[str, Any]):
        """Add snakemake args to config."""
        config["snakemake_args"] = args

    @bidsapp.hookimpl
    def update_cli_namespace(self, namespace: dict[str, Any], config: dict[str, Any]):
        """Resolve Paths within the namespace and set target."""
        if "output_dir" in namespace:
            namespace["output_dir"] = Path(namespace["output_dir"])
        if "bids_dir" in namespace:
            namespace["bids_dir"] = Path(namespace["bids_dir"])
        for key, path in namespace.items():
            namespace[key] = _resolve_path(path)
        namespace.pop("help_snakemake", None)
        self.force_output = namespace.pop("force_output", False)
        if "targets_by_analysis_level" in config and "analysis_level" in namespace:
            config["snakemake_target"] = config["targets_by_analysis_level"][
                namespace["analysis_level"]
            ]
        else:
            config["snakemake_target"] = None

    @bidsapp.hookimpl
    def finalize_config(self, config: dict[str, Any]):
        """Perform final steps for snakemake workflows.

        Expects to find ``output_dir`` in config as a fully resolved
        :class:`~pathlib.Path`

        - Modify ``output_dir`` to `output/results` if output is set to the app dir
        - Set ``self.cwd``
        - Write config file into output dir
        - Add snakemake specific parameters to config

        """
        # First, handle outputs in snakebids_root or results folder
        if config["output_dir"] == self.snakemake_dir.resolve():
            config["output_dir"] /= "results"
            # Print a friendly warning that the output directory will change
            logger.info(
                "You specified your output to be in the snakebids directory, so "
                "we're automatically putting your outputs in the results "
                "subdirectory.\nYou'll find your results in `%s`",
                (self.snakemake_dir / "results").resolve(),
            )

        if config["output_dir"].is_relative_to(self.snakemake_dir / "results"):
            write_output_mode(self.snakemake_dir / ".snakebids", "workflow")
            self.cwd = self.snakemake_dir
            root = config["output_dir"].relative_to(self.snakemake_dir)
        # else, we run in bidsapp mode
        else:
            try:
                prepare_bidsapp_output(config["output_dir"], self.force_output)
            except RunError as err:
                print(err.msg, file=sys.stderr)
                sys.exit(1)
            self.cwd = config["output_dir"]
            root = ""  # uses "" instead of Path() to drop the leading ./

        configfile_path = self.configfile_path or self.snakemake_dir / "snakebids.yaml"
        if self.configfile_outpath is None:
            try:
                self.configfile_outpath = (
                    self.cwd
                    / configfile_path.resolve().relative_to(
                        self.snakemake_dir.resolve()
                    )
                )
            except ValueError:
                self.configfile_outpath = configfile_path

        version = config.get("plugins.version.version")
        if version == "unknown":
            logger.warning(
                "App version could not be identified. This usually occurs because the "
                "app was not installed with pip. It will be recorded in output as "
                "'unknown'."
            )
        # Write the config file
        write_config(
            config_file=self.configfile_outpath,
            data=dict(
                config,
                snakemake_version=impm.version("snakemake"),
                snakebids_version=impm.version("snakebids"),
                root=root,
                app_version=version or "unknown",
                snakemake_dir=self.snakemake_dir,
                snakefile=self.snakefile_path,
            ),
            force_overwrite=True,
        )

    @bidsapp.hookimpl
    def run(self, config: dict[str, Any]):
        """Run snakemake with the given config, after applying plugins."""
        snakemake_main(  # type: ignore
            [
                *filter(
                    None,
                    [
                        *itx.always_iterable(config["snakemake_target"]),
                        "--snakefile",
                        str(self.snakefile_path),
                        "--directory",
                        str(self.cwd),
                        "--configfile",
                        str(self.configfile_outpath),
                        *config["snakemake_args"],
                    ],
                )
            ]
        )



# type: ignore
__submodules__ = [
    "bidsargs",
    "cli_config",
    "component_edit",
    "pybidsdb",
    "validator",
    "snakemake",
    "version",
]

# <AUTOGEN_INIT>
import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

__all__ = [
    "BidsArgs",
    "BidsValidator",
    "CliConfig",
    "ComponentEdit",
    "FilterParse",
    "FilterParseError",
    "InvalidBidsError",
    "Pybidsdb",
    "SnakemakeBidsApp",
    "Version",
    "logger",
]
# </AUTOGEN_INIT>



"""
This type stub file was generated by pyright.
"""

import io
import os
import sys
from typing import Any, AnyStr, Optional, Union, overload

"""Helper classes use for fake file system implementation."""
AnyString = Union[str, bytes]
AnyPath = Union[AnyStr, "os.PathLike[AnyStr]"]
IS_PYPY = ...
IS_WIN = ...
IN_DOCKER = ...
PERM_READ = ...
PERM_WRITE = ...
PERM_EXE = ...
PERM_DEF = ...
PERM_DEF_FILE = ...
PERM_ALL = ...
if sys.platform == "win32": ...
else:
    USER_ID = ...
    GROUP_ID = ...

def set_uid(uid: int) -> None:
    """Set the global user id. This is used as st_uid for new files
    and to differentiate between a normal user and the root user (uid 0).
    For the root user, some permission restrictions are ignored.

    Args:
        uid: (int) the user ID of the user calling the file system functions.
    """
    ...

def set_gid(gid: int) -> None:
    """Set the global group id. This is only used to set st_gid for new files,
    no permission checks are performed.

    Args:
        gid: (int) the group ID of the user calling the file system functions.
    """
    ...

def reset_ids() -> None:
    """Set the global user ID and group ID back to default values."""
    ...

def is_root() -> bool:
    """Return True if the current user is the root user."""
    ...

def is_int_type(val: Any) -> bool:
    """Return True if `val` is of integer type."""
    ...

def is_byte_string(val: Any) -> bool:
    """Return True if `val` is a bytes-like object, False for a unicode
    string.
    """
    ...

def is_unicode_string(val: Any) -> bool:
    """Return True if `val` is a unicode string, False for a bytes-like
    object.
    """
    ...

@overload
def make_string_path(dir_name: AnyStr) -> AnyStr: ...
@overload
def make_string_path(dir_name: os.PathLike) -> str: ...
def make_string_path(dir_name: AnyPath) -> AnyStr: ...
def to_string(path: Union[AnyStr, Union[str, bytes]]) -> str:
    """Return the string representation of a byte string using the preferred
    encoding, or the string itself if path is a str.
    """
    ...

def to_bytes(path: Union[AnyStr, Union[str, bytes]]) -> bytes:
    """Return the bytes representation of a string using the preferred
    encoding, or the byte string itself if path is a byte string.
    """
    ...

def join_strings(s1: AnyStr, s2: AnyStr) -> AnyStr:
    """This is a bit of a hack to satisfy mypy - may be refactored."""
    ...

def real_encoding(encoding: Optional[str]) -> Optional[str]:
    """Since Python 3.10, the new function ``io.text_encoding`` returns
    "locale" as the encoding if None is defined. This will be handled
    as no encoding in pyfakefs.
    """
    ...

def now(): ...
@overload
def matching_string(matched: bytes, string: AnyStr) -> bytes: ...
@overload
def matching_string(matched: str, string: AnyStr) -> str: ...
@overload
def matching_string(matched: AnyStr, string: None) -> None: ...
def matching_string(matched: AnyStr, string: Optional[AnyStr]) -> Optional[AnyString]:
    """Return the string as byte or unicode depending
    on the type of matched, assuming string is an ASCII string.
    """
    ...

class FakeStatResult:
    """Mimics os.stat_result for use as return type of `stat()` and similar.
    This is needed as `os.stat_result` has no possibility to set
    nanosecond times directly.
    """

    def __init__(
        self,
        is_windows: bool,
        user_id: int,
        group_id: int,
        initial_time: Optional[float] = ...,
    ) -> None: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def copy(self) -> FakeStatResult:
        """Return a copy where the float usage is hard-coded to mimic the
        behavior of the real os.stat_result.
        """
        ...
    def set_from_stat_result(self, stat_result: os.stat_result) -> None:
        """Set values from a real os.stat_result.
        Note: values that are controlled by the fake filesystem are not set.
        This includes st_ino, st_dev and st_nlink.
        """
        ...
    @property
    def st_ctime(self) -> Union[int, float]:
        """Return the creation time in seconds."""
        ...
    @st_ctime.setter
    def st_ctime(self, val: Union[int, float]) -> None:
        """Set the creation time in seconds."""
        ...
    @property
    def st_atime(self) -> Union[int, float]:
        """Return the access time in seconds."""
        ...
    @st_atime.setter
    def st_atime(self, val: Union[int, float]) -> None:
        """Set the access time in seconds."""
        ...
    @property
    def st_mtime(self) -> Union[int, float]:
        """Return the modification time in seconds."""
        ...
    @st_mtime.setter
    def st_mtime(self, val: Union[int, float]) -> None:
        """Set the modification time in seconds."""
        ...
    @property
    def st_size(self) -> int: ...
    @st_size.setter
    def st_size(self, val: int) -> None: ...
    @property
    def st_blocks(self) -> int:
        """Return the number of 512-byte blocks allocated for the file.
        Assumes a page size of 4096 (matches most systems).
        Ignores that this may not be available under some systems,
        and that the result may differ if the file has holes.
        """
        ...
    @property
    def st_file_attributes(self) -> int: ...
    @property
    def st_reparse_tag(self) -> int: ...
    def __getitem__(self, item: int) -> Optional[int]:
        """Implement item access to mimic `os.stat_result` behavior."""
        ...
    @property
    def st_atime_ns(self) -> int:
        """Return the access time in nanoseconds."""
        ...
    @st_atime_ns.setter
    def st_atime_ns(self, val: int) -> None:
        """Set the access time in nanoseconds."""
        ...
    @property
    def st_mtime_ns(self) -> int:
        """Return the modification time in nanoseconds."""
        ...
    @st_mtime_ns.setter
    def st_mtime_ns(self, val: int) -> None:
        """Set the modification time of the fake file in nanoseconds."""
        ...
    @property
    def st_ctime_ns(self) -> int:
        """Return the creation time in nanoseconds."""
        ...
    @st_ctime_ns.setter
    def st_ctime_ns(self, val: int) -> None:
        """Set the creation time of the fake file in nanoseconds."""
        ...

class BinaryBufferIO(io.BytesIO):
    """Stream class that handles byte contents for files."""

    def __init__(self, contents: Optional[bytes]) -> None: ...
    def putvalue(self, value: bytes) -> None: ...

class TextBufferIO(io.TextIOWrapper):
    """Stream class that handles Python string contents for files."""

    def __init__(
        self,
        contents: Optional[bytes] = ...,
        newline: Optional[str] = ...,
        encoding: Optional[str] = ...,
        errors: str = ...,
    ) -> None: ...
    def getvalue(self) -> bytes: ...
    def putvalue(self, value: bytes) -> None: ...



"""
This type stub file was generated by pyright.
"""

if False: ...
__all__ = ["ScalarBoolean"]

class ScalarBoolean(int):
    def __new__(cls: Any, *args: Any, **kw: Any) -> Any: ...
    @property
    def anchor(self) -> Any: ...
    def yaml_anchor(self, any: bool = ...) -> Any: ...
    def yaml_set_anchor(self, value: Any, always_dump: bool = ...) -> None: ...



"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.error import YAMLError

if False: ...
__all__ = ["Reader", "ReaderError"]

class ReaderError(YAMLError):
    def __init__(
        self, name: Any, position: Any, character: Any, encoding: Any, reason: Any
    ) -> None: ...
    def __str__(self) -> Any: ...

class Reader:
    def __init__(self, stream: Any, loader: Any = ...) -> None: ...
    def reset_reader(self) -> None: ...
    @property
    def stream(self) -> Any: ...
    @stream.setter
    def stream(self, val: Any) -> None: ...
    def peek(self, index: int = ...) -> Text: ...
    def prefix(self, length: int = ...) -> Any: ...
    def forward_1_1(self, length: int = ...) -> None: ...
    def forward(self, length: int = ...) -> None: ...
    def get_mark(self) -> Any: ...
    def determine_encoding(self) -> None: ...

    NON_PRINTABLE = ...
    _printable_ascii = ...
    def check_printable(self, data: Any) -> None: ...
    def update(self, length: int) -> None: ...
    def update_raw(self, size: Optional[int] = ...) -> None: ...



from __future__ import annotations

import itertools as it
import json
import operator as op
import textwrap
from collections.abc import Sequence
from math import ceil, floor, inf

import more_itertools as itx

from snakebids.types import ZipList


def quote_wrap(val: str) -> str:
    """Wrap string in quotes, with additional character escapes for printing."""
    return (
        json.dumps(val, ensure_ascii=False)
        .replace("\x85", "\\x85")
        .replace("\u2028", "\\u2028")
        .replace("\u2029", "\\u2029")
    )


def format_zip_lists(
    zip_list: ZipList, max_width: int | float | None = None, tabstop: int = 4
) -> str:
    """Pretty-format zip-lists in a tablar format.

    Parameters
    ----------
    zip_list
        zip_list to format
    max_width
        Maximum character width. If possible, zip_list values will be elided to fit
        within this width
    tabstop
        Number of spaces to include in each level of indentation
    """
    if not zip_list:
        return "{}"
    table = [_format_zip_row(key, row) for key, row in zip_list.items()]
    widths = [max(len(val) for val in col) for col in zip(*table, strict=True)]
    aligned = _align_zip_table(table, widths)
    cols: list[Sequence[str]] = list(zip(*aligned, strict=True))
    keys = cols[:2]
    vals = cols[2:]
    elided_cols = it.chain(
        keys,
        _elide_zip_table(
            vals,
            len(aligned),
            widths[2:],
            # max width reduces for the indent, minus 2 for the closing "]," minus the
            # width of the key and opening brace
            max_width=(max_width or inf) - tabstop - 2 - sum(widths[:2]),
        ),
        [["],\n"] * len(table)],
    )
    return "".join(
        [
            "{\n",
            textwrap.indent(
                "".join(itx.flatten(zip(*elided_cols, strict=True))), " " * tabstop
            ),
            "}",
        ]
    )


def _format_zip_row(key: str, row: list[str]) -> list[str]:
    formatted_values = [
        quote_wrap(val) + sep
        for val, sep in it.zip_longest(row, it.repeat(", ", len(row) - 1), fillvalue="")
    ]
    row = [f"{quote_wrap(key)}: "]
    row.append("[")
    row.extend(formatted_values)
    return row


def _align_zip_table(table: list[list[str]], widths: list[int]) -> list[list[str]]:
    output: list[list[str]] = []
    for row in table:
        spaces = [
            " " * (width - len(val)) for val, width in zip(row, widths, strict=True)
        ]
        output.append(list(map(op.add, row, spaces)))
    return output


def _elide_zip_table(
    cols: list[Sequence[str]], col_len: int, widths: list[int], max_width: int | float
) -> list[Sequence[str]]:
    def new_col(val: str):
        return [[val] * col_len]

    overflow = int(max(sum(widths) - max_width, 0))
    elision = _find_elision(list(widths), slice(0, 0), overflow)
    if elision != slice(0, 0):
        return list(
            it.chain(
                cols[: elision.start],
                new_col("..."),
                (new_col(" ") if elision.stop - elision.start < len(cols) else []),
                cols[elision.stop :],
            )
        )
    return cols


def _find_elision(widths: list[int], excluded: slice, overflow: int) -> slice:
    # Add 4 to overflow to account for ellipses
    if max(sum(widths[excluded]) - 4, 0) >= overflow:
        return excluded
    span = excluded.stop - excluded.start
    # don't let span become longer than list of values
    if span >= len(widths):
        return excluded

    num_vals = len(widths)
    mid = floor(num_vals / 2)

    # need different rules for handling exclusions of even length depending on whether
    # there's an even or odd total number of values.
    left_bias = floor if num_vals % 2 else ceil
    right_bias = ceil if num_vals % 2 else floor

    return _find_elision(
        widths,
        slice(mid - left_bias(span / 2), mid + right_bias(span / 2) + 1, 1),
        overflow,
    )



from __future__ import annotations

import functools as ft
from typing import TYPE_CHECKING

from snakebids.paths._templates import spec_func
from snakebids.paths._utils import BidsPathSpec, find_entity, get_spec_path, load_spec

# <AUTOUPDATE>
# The code between these tags is automatically generated. Do not
# manually edit
# To update, run::
#
#       poetry run poe update-bids
#
if not TYPE_CHECKING:
    __all__ = [  # noqa:F822
        "LATEST",
        "latest",
        "v0_0_0",
        "v0_11_0",
        "v0_15_0",
    ]

    def __dir__():
        return __all__


_SPECS = ["v0_0_0", "v0_11_0", "v0_15_0"]
# LATEST = "v0_15_0"
# </AUTOUPDATE>

# To automatically use latest spec as "LATEST", remove this line and uncomment the
# generated line in scripts/update_bids.py
LATEST = "v0_0_0"


@ft.lru_cache
def __getattr__(name: str):
    """Allow dynamic retrieval of latest spec."""
    if name == "latest":
        name = LATEST

    if name not in _SPECS:
        msg = f"module '{__name__}' has no attribute '{name}'"
        raise AttributeError(msg)

    spec_config = load_spec(get_spec_path(name))

    spec = spec_config["spec"]

    def get_spec(subject_dir: bool = True, session_dir: bool = True) -> BidsPathSpec:
        if not subject_dir:
            find_entity(spec, "subject")["dir"] = False

        if not session_dir:
            find_entity(spec, "session")["dir"] = False

        return spec

    get_spec.__doc__ = spec_func.format_doc(spec_config)
    get_spec.__name__ = name

    return get_spec



"""
This type stub file was generated by pyright.
"""
from __future__ import annotations

import enum
from functools import lru_cache
from pathlib import Path
from typing import Any, Callable, Sequence

from bids import BIDSLayoutIndexer
from bids.layout.models import BIDSFile, Entity

"""BIDSLayout class."""
__all__ = ["BIDSLayout"]

class BIDSLayout:
    """Layout class representing an entire BIDS dataset.

    Parameters
    ----------
    root : str
        The root directory of the BIDS dataset.
    validate : bool, optional
        If True, all files are checked for BIDS compliance when first indexed,
        and non-compliant files are ignored. This provides a convenient way to
        restrict file indexing to only those files defined in the "core" BIDS
        spec, as setting validate=True will lead files in supplementary folders
        like derivatives/, code/, etc. to be ignored.
    absolute_paths : bool, optional
        If True, queries always return absolute paths.
        If False, queries return relative paths (for files and
        directories).
    derivatives : bool or str or list, optional
        Specifies whether and/or which
        derivatives to index. If True, all pipelines found in the
        derivatives/ subdirectory will be indexed. If a str or list, gives
        the paths to one or more derivatives directories to index. If False
        or None, the derivatives/ directory is ignored during indexing, and
        derivatives will have to be added manually via add_derivatives().
        Note: derivatives datasets MUST contain a dataset_description.json
        file in order to be indexed.
    config : str or list or None, optional
        Optional name(s) of configuration file(s) to use.
        By default (None), uses 'bids'.
    sources : :obj:`bids.layout.BIDSLayout` or list or None, optional
        Optional BIDSLayout(s) from which the current BIDSLayout is derived.
    config_filename : str
        Optional name of filename within directories
        that contains configuration information.
    regex_search : bool
        Whether to require exact matching (True) or regex
        search (False, default) when comparing the query string to each
        entity in .get() calls. This sets a default for the instance, but
        can be overridden in individual .get() requests.
    database_path : str
        Optional path to directory containing SQLite database file index
        for this BIDS dataset. If a value is passed and the folder
        already exists, indexing is skipped. By default (i.e., if None),
        an in-memory SQLite database is used, and the index will not
        persist unless .save() is explicitly called.
    reset_database : bool
        If True, any existing directory specified in the
        database_path argument is deleted, and the BIDS dataset provided
        in the root argument is reindexed. If False, indexing will be
        skipped and the existing database file will be used. Ignored if
        database_path is not provided.
    indexer: BIDSLayoutIndexer or callable
        An optional BIDSLayoutIndexer instance to use for indexing, or any
        callable that takes a BIDSLayout instance as its only argument. If
        None, a new indexer with default parameters will be implicitly created.
    indexer_kwargs: dict
        Optional keyword arguments to pass onto the newly created
        BIDSLayoutIndexer. Valid keywords are 'ignore', 'force_index',
        'index_metadata', and 'config_filename'. Ignored if indexer is not
        None.
    """

    def __init__(
        self,
        root: str = ...,
        validate: bool = ...,
        absolute_paths: bool = ...,
        derivatives: bool | str | Path | list[str] | None = ...,
        config: str | Path | list[str] | None = ...,
        sources: BIDSLayout | list[BIDSLayout] | None = ...,
        regex_search: bool = ...,
        database_path: str | Path | None = ...,
        reset_database: bool = ...,
        indexer: BIDSLayoutIndexer | Callable[[BIDSLayout], Any] = ...,
        **indexer_kwargs: dict[str, bool | str],
    ) -> None: ...
    @property
    def root(self) -> str: ...
    def __repr__(self) -> str:
        """Provide a tidy summary of key properties."""
        ...
    @property
    def session(self): ...
    @property
    @lru_cache()
    def config(self): ...
    @property
    def entities(self) -> dict[str, Entity]:
        """Get the entities."""
        ...
    @property
    def files(self):  # -> dict[Unknown, Unknown]:
        """Get the files."""
        ...
    @classmethod
    def load(cls, database_path):  # -> Self@BIDSLayout:
        """Load index from database path. Initialization parameters are set to
        those found in database_path JSON sidecar.

        Parameters
        ----------
        database_path : str, Path
            The path to the desired database folder. If a relative path is
            passed, it is assumed to be relative to the BIDSLayout root
            directory.
        """
        ...
    def save(self, database_path, replace_connection=...):  # -> None:
        """Save the current index as a SQLite3 DB at the specified location.

        Note: This is only necessary if a database_path was not specified
        at initialization, and the user now wants to save the index.
        If a database_path was specified originally, there is no need to
        re-save using this method.

        Parameters
        ----------
        database_path : str
            The path to the desired database folder. By default,
            uses .db_cache. If a relative path is passed, it is assumed to
            be relative to the BIDSLayout root directory.
        replace_connection : bool, optional
            If True, the newly created database will
            be used for all subsequent connections. This means that any
            changes to the index made after the .save() call will be
            reflected in the database file. If False, the previous database
            will continue to be used, and any subsequent changes will not
            be reflected in the new file unless save() is explicitly called
            again.
        """
        ...
    def get_entities(self, scope=..., metadata=...):  # -> dict[Unknown, Unknown]:
        """Get entities for all layouts in the specified scope.

        Parameters
        ----------
        scope : str
            The scope of the search space. Indicates which
            BIDSLayouts' entities to extract.
            See :obj:`bids.layout.BIDSLayout.get` docstring for valid values.

        metadata : bool or None
            By default (None), all available entities
            are returned. If True, only entities found in metadata files
            (and not defined for filenames) are returned. If False, only
            entities defined for filenames (and not those found in JSON
            sidecars) are returned.

        Returns
        -------
        dict
            Dictionary where keys are entity names and
            values are Entity instances.
        """
        ...
    def get_files(self, scope=...):  # -> dict[Unknown, Unknown]:
        """Get BIDSFiles for all layouts in the specified scope.

        Parameters
        ----------
        scope : str
            The scope of the search space. Indicates which
            BIDSLayouts' entities to extract.
            See :obj:`bids.layout.BIDSLayout.get` docstring for valid values.


        Returns:
            A dict, where keys are file paths and values
            are :obj:`bids.layout.BIDSFile` instances.

        """
        ...
    def clone(self):  # -> Self@BIDSLayout:
        """Return a deep copy of the current BIDSLayout."""
        ...
    def parse_file_entities(
        self, filename, scope=..., entities=..., config=..., include_unmatched=...
    ):  # -> dict[Unknown, Unknown]:
        """Parse the passed filename for entity/value pairs.

        Parameters
        ----------
        filename : str
            The filename to parse for entity values
        scope : str or list, optional
            The scope of the search space. Indicates which BIDSLayouts'
            entities to extract. See :obj:`bids.layout.BIDSLayout.get`
            docstring for valid values. By default, extracts all entities.
        entities : list or None, optional
            An optional list of Entity instances to use in
            extraction. If passed, the scope and config arguments are
            ignored, and only the Entities in this list are used.
        config : str or :obj:`bids.layout.models.Config` or list or None, optional
            One or more :obj:`bids.layout.models.Config` objects, or paths
            to JSON config files on disk, containing the Entity definitions
            to use in extraction. If passed, scope is ignored.
        include_unmatched : bool, optional
            If True, unmatched entities are included
            in the returned dict, with values set to None. If False
            (default), unmatched entities are ignored.

        Returns
        -------
        dict
            Dictionary where keys are Entity names and values are the
            values extracted from the filename.
        """
        ...
    def add_derivatives(self, path, parent_database_path=..., **kwargs):  # -> None:
        """Add BIDS-Derivatives datasets to tracking.

        Parameters
        ----------
        path : str or list
            One or more paths to BIDS-Derivatives datasets.
            Each path can point to either a derivatives/ directory
            containing one more more pipeline directories, or to a single
            pipeline directory (e.g., derivatives/fmriprep).
        parent_database_path : str or Path
            If not None, use the pipeline name from the dataset_description.json
            file as the database folder name to nest within the parent database
            folder name to write out derivative index to.
        kwargs : dict
            Optional keyword arguments to pass on to
            BIDSLayout() when initializing each of the derivative datasets.

        Notes
        -----
        Every derivatives directory intended for indexing MUST contain a
        valid dataset_description.json file. See the BIDS-Derivatives
        specification for details.
        """
        ...
    def to_df(self, metadata=..., **filters):
        """Return information for BIDSFiles tracked in Layout as pd.DataFrame.

        Parameters
        ----------
        metadata : bool, optional
            If True, includes columns for all metadata fields.
            If False, only filename-based entities are included as columns.
        filters : dict, optional
            Optional keyword arguments passed on to get(). This allows
            one to easily select only a subset of files for export.

        Returns
        -------
        :obj:`pandas.DataFrame`
            A pandas DataFrame, where each row is a file, and each column is
            a tracked entity. NaNs are injected whenever a file has no
            value for a given attribute.

        """
        ...
    def get(
        self,
        # return_type=...,
        # target=...,
        # scope=...,
        regex_search: bool = ...,
        # absolute_paths=...,
        # invalid_filters=...,
        **filters: str | Query | Sequence[str | Query],
    ) -> list[BIDSFile]:
        """Retrieve files and/or metadata from the current Layout.

        Parameters
        ----------
        return_type : str, optional
            Type of result to return. Valid values:
            'object' (default): return a list of matching BIDSFile objects.
            'file' or 'filename': return a list of matching filenames.
            'dir': return a list of directories.
            'id': return a list of unique IDs. Must be used together
                  with a valid target.
        target : str, optional
            Optional name of the target entity to get results for
            (only used if return_type is 'dir' or 'id').
        scope : str or list, optional
            Scope of the search space. If passed, only
            nodes/directories that match the specified scope will be
            searched. Possible values include:
            'all' (default): search all available directories.
            'derivatives': search all derivatives directories.
            'raw': search only BIDS-Raw directories.
            'self': search only the directly called BIDSLayout.
            <PipelineName>: the name of a BIDS-Derivatives pipeline.
        regex_search : bool or None, optional
            Whether to require exact matching
            (False) or regex search (True) when comparing the query string
            to each entity.
        absolute_paths : bool, optional
            Optionally override the instance-wide option
            to report either absolute or relative (to the top of the
            dataset) paths. If None, will fall back on the value specified
            at BIDSLayout initialization.
        invalid_filters (str): Controls behavior when named filters are
            encountered that don't exist in the database (e.g., in the case of
            a typo like subbject='0.1'). Valid values:
                'error' (default): Raise an explicit error.
                'drop': Silently drop invalid filters (equivalent to not having
                    passed them as arguments in the first place).
                'allow': Include the invalid filters in the query, resulting
                    in no results being returned.
        filters : dict
            Any optional key/values to filter the entities on.
            Keys are entity names, values are regexes to filter on. For
            example, passing filters={'subject': 'sub-[12]'} would return
            only files that match the first two subjects. In addition to
            ordinary data types, the following enums are defined (in the
            Query class):
                * Query.NONE: The named entity must not be defined.
                * Query.ANY: the named entity must be defined, but can have any
                    value.

        Returns
        -------
        list of :obj:`bids.layout.BIDSFile` or str
            A list of BIDSFiles (default) or strings (see return_type).
        """
        ...
    def get_file(self, filename, scope=...):  # -> None:
        """Return the BIDSFile object with the specified path.

        Parameters
        ----------
        filename : str
            The path of the file to retrieve. Must be either an absolute path,
            or relative to the root of this BIDSLayout.
        scope : str or list, optional
            Scope of the search space. If passed, only BIDSLayouts that match
            the specified scope will be searched. See :obj:`BIDSLayout.get`
            docstring for valid values. Default is 'all'.

        Returns
        -------
        :obj:`bids.layout.BIDSFile` or None
            File found, or None if no match was found.
        """
        ...
    def get_collections(
        self,
        level,
        types=...,
        variables=...,
        merge=...,
        sampling_rate=...,
        skip_empty=...,
        **kwargs,
    ):
        """Return one or more variable Collections in the BIDS project.

        Parameters
        ----------
        level : {'run', 'session', 'subject', 'dataset'}
            The level of analysis to return variables for.
            Must be one of 'run', 'session','subject', or 'dataset'.
        types : str or list
            Types of variables to retrieve. All valid values reflect the
            filename stipulated in the BIDS spec for each kind of variable.
            Valid values include: 'events', 'physio', 'stim', 'scans',
            'participants', 'sessions', and 'regressors'. Default is None.
        variables : list
            Optional list of variables names to return. If None, all available
            variables are returned.
        merge : bool
            If True, variables are merged across all observations of the
            current level. E.g., if level='subject', variables from all
            subjects will be merged into a single collection. If False, each
            observation is handled separately, and the result is returned
            as a list.
        sampling_rate : int or str
            If level='run', the sampling rate to pass onto the returned
            :obj:`bids.variables.collections.BIDSRunVariableCollection`.
        skip_empty : bool
            Whether or not to skip empty Variables (i.e., where there are no
            rows/records in a file after applying any filtering operations
            like dropping NaNs).
        kwargs
            Optional additional arguments to pass onto
            :obj:`bids.variables.io.load_variables`.

        Returns
        -------
        list of :obj:`bids.variables.collections.BIDSVariableCollection`
            or :obj:`bids.variables.collections.BIDSVariableCollection`
            A list if merge=False;
            a single :obj:`bids.variables.collections.BIDSVariableCollection`
            if merge=True.

        """
        ...
    def get_metadata(self, path, include_entities=..., scope=...):  # -> BIDSMetadata:
        """Return metadata found in JSON sidecars for the specified file.

        Parameters
        ----------
        path : str
            Path to the file to get metadata for.
        include_entities : bool, optional
            If True, all available entities extracted
            from the filename (rather than JSON sidecars) are included in
            the returned metadata dictionary.
        scope : str or list, optional
            The scope of the search space. Each element must
            be one of 'all', 'raw', 'self', 'derivatives', or a
            BIDS-Derivatives pipeline name. Defaults to searching all
            available datasets.

        Returns
        -------
        dict
            A dictionary of key/value pairs extracted from all of the
            target file's associated JSON sidecars.

        Notes
        -----
        A dictionary containing metadata extracted from all matching .json
        files is returned. In cases where the same key is found in multiple
        files, the values in files closer to the input filename will take
        precedence, per the inheritance rules in the BIDS specification.

        """
        ...
    def get_dataset_description(self, scope=..., all_=...):  # -> list[Unknown]:
        """Return contents of dataset_description.json.

        Parameters
        ----------
        scope : str
            The scope of the search space. Only descriptions of
            BIDSLayouts that match the specified scope will be returned.
            See :obj:`bids.layout.BIDSLayout.get` docstring for valid values.
            Defaults to 'self' --i.e., returns the dataset_description.json
            file for only the directly-called BIDSLayout.
        all_ : bool
            If True, returns a list containing descriptions for
            all matching layouts. If False (default), returns for only the
            first matching layout.

        Returns
        -------
        dict or list of dict
            a dictionary or list of dictionaries (depending on all_).
        """
        ...
    def get_nearest(
        self,
        path,
        return_type=...,
        strict=...,
        all_=...,
        ignore_strict_entities=...,
        full_search=...,
        **filters,
    ):  # -> list[Unknown] | None:
        """Walk up file tree from specified path and return nearest matching file(s).

        Parameters
        ----------
        path (str): The file to search from.
        return_type (str): What to return; must be one of 'filename'
            (default) or 'tuple'.
        strict (bool): When True, all entities present in both the input
            path and the target file(s) must match perfectly. When False,
            files will be ordered by the number of matching entities, and
            partial matches will be allowed.
        all_ (bool): When True, returns all matching files. When False
            (default), only returns the first match.
        ignore_strict_entities (str, list): Optional entity/entities to
            exclude from strict matching when strict is True. This allows
            one to search, e.g., for files of a different type while
            matching all other entities perfectly by passing
            ignore_strict_entities=['type']. Ignores extension by default.
        full_search (bool): If True, searches all indexed files, even if
            they don't share a common root with the provided path. If
            False, only files that share a common root will be scanned.
        filters : dict
            Optional keywords to pass on to :obj:`bids.layout.BIDSLayout.get`.
        """
        ...
    def get_bvec(self, path, **kwargs):
        """Get bvec file for passed path."""
        ...
    def get_bval(self, path, **kwargs):
        """Get bval file for passed path."""
        ...
    def get_fieldmap(self, path, return_list=...):  # -> list[Unknown] | None:
        """Get fieldmap(s) for specified path."""
        ...
    def get_tr(self, derivatives=..., **filters):
        """Return the scanning repetition time (TR) for one or more runs.

        Parameters
        ----------
        derivatives : bool
            If True, also checks derivatives images.
        filters : dict
            Optional keywords used to constrain the selected runs.
            Can be any arguments valid for a .get call (e.g., BIDS entities
            or JSON sidecar keys).

        Returns
        -------
        float
            A single float.

        Notes
        -----
        Raises an exception if more than one unique TR is found.
        """
        ...
    def build_path(
        self,
        source,
        path_patterns=...,
        strict=...,
        scope=...,
        validate=...,
        absolute_paths=...,
    ):  # -> str:
        """Construct a target filename for a file or dictionary of entities.

        Parameters
        ----------
        source : str or :obj:`bids.layout.BIDSFile` or dict
            The source data to use to construct the new file path.
            Must be one of:
            - A BIDSFile object
            - A string giving the path of a BIDSFile contained within the
              current Layout.
            - A dict of entities, with entity names in keys and values in
              values
        path_patterns : list
            Optional path patterns to use to construct
            the new file path. If None, the Layout-defined patterns will
            be used. Entities should be represented by the name
            surrounded by curly braces. Optional portions of the patterns
            should be denoted by square brackets. Entities that require a
            specific value for the pattern to match can pass them inside
            angle brackets. Default values can be assigned by specifying a string
            after the pipe operator. E.g., (e.g., {type<image>|bold} would
            only match the pattern if the entity 'type' was passed and its
            value is "image", otherwise the default value "bold" will be
            used).
                Example: 'sub-{subject}/[var-{name}/]{id}.csv'
                Result: 'sub-01/var-SES/1045.csv'
        strict : bool, optional
            If True, all entities must be matched inside a
            pattern in order to be a valid match. If False, extra entities
            will be ignored so long as all mandatory entities are found.
        scope : str or list, optional
            The scope of the search space. Indicates which
            BIDSLayouts' path patterns to use. See BIDSLayout docstring
            for valid values. By default, uses all available layouts. If
            two or more values are provided, the order determines the
            precedence of path patterns (i.e., earlier layouts will have
            higher precedence).
        validate : bool, optional
            If True, built path must pass BIDS validator. If
            False, no validation is attempted, and an invalid path may be
            returned (e.g., if an entity value contains a hyphen).
        absolute_paths : bool, optional
            Optionally override the instance-wide option
            to report either absolute or relative (to the top of the
            dataset) paths. If None, will fall back on the value specified
            at BIDSLayout initialization.
        """
        ...
    def copy_files(
        self,
        files=...,
        path_patterns=...,
        symbolic_links=...,
        root=...,
        conflicts=...,
        **kwargs,
    ):  # -> None:
        """Copy BIDSFile(s) to new locations.

        The new locations are defined by each BIDSFile's entities and the
        specified `path_patterns`.

        Parameters
        ----------
        files : list
            Optional list of BIDSFile objects to write out. If
            none provided, use files from running a get() query using
            remaining **kwargs.
        path_patterns : str or list
            Write patterns to pass to each file's write_file method.
        symbolic_links : bool
            Whether to copy each file as a symbolic link or a deep copy.
        root : str
            Optional root directory that all patterns are relative
            to. Defaults to dataset root.
        conflicts : str
            Defines the desired action when the output path already exists.
            Must be one of:
                'fail': raises an exception
                'skip' does nothing
                'overwrite': overwrites the existing file
                'append': adds a suffix to each file copy, starting with 1
        kwargs : dict
            Optional key word arguments to pass into a get() query.
        """
        ...
    def write_to_file(
        self,
        entities,
        path_patterns=...,
        contents=...,
        link_to=...,
        copy_from=...,
        content_mode=...,
        conflicts=...,
        strict=...,
        validate=...,
    ):  # -> None:
        """Write data to a file defined by the passed entities and patterns.

        Parameters
        ----------
        entities : dict
            A dictionary of entities, with Entity names in
            keys and values for the desired file in values.
        path_patterns : list
            Optional path patterns to use when building
            the filename. If None, the Layout-defined patterns will be
            used.
        contents : object
            Contents to write to the generate file path.
            Can be any object serializable as text or binary data (as
            defined in the content_mode argument).
        link_to : str
            Optional path with which to create a symbolic link
            to. Used as an alternative to and takes priority over the
            contents argument.
        conflicts : str
            Defines the desired action when the output path already exists.
            Must be one of:
                'fail': raises an exception
                'skip' does nothing
                'overwrite': overwrites the existing file
                'append': adds a suffix to each file copy, starting with 1
        strict : bool
            If True, all entities must be matched inside a
            pattern in order to be a valid match. If False, extra entities
            will be ignored so long as all mandatory entities are found.
        validate : bool
            If True, built path must pass BIDS validator. If
            False, no validation is attempted, and an invalid path may be
            returned (e.g., if an entity value contains a hyphen).
        """
        ...

class Query(enum.Enum):
    """Enums for use with BIDSLayout.get()."""

    NONE = ...
    ANY = ...
    OPTIONAL = ...



"""
This type stub file was generated by pyright.
"""

from .error import StreamMark

if False: ...
SHOW_LINES = ...

class Token:
    __slots__ = ...
    def __init__(self, start_mark: StreamMark, end_mark: StreamMark) -> None: ...
    def __repr__(self) -> Any: ...
    @property
    def column(self) -> int: ...
    @column.setter
    def column(self, pos: Any) -> None: ...
    def add_post_comment(self, comment: Any) -> None: ...
    def add_pre_comments(self, comments: Any) -> None: ...
    def add_comment_pre(self, comment: Any) -> None: ...
    def add_comment_eol(self, comment: Any, comment_type: Any) -> None: ...
    def add_comment_post(self, comment: Any) -> None: ...
    @property
    def comment(self) -> Any: ...
    def move_old_comment(self, target: Any, empty: bool = ...) -> Any:
        """move a comment from this token to target (normally next token)
        used to combine e.g. comments before a BlockEntryToken to the
        ScalarToken that follows it
        empty is a special for empty values -> comment after key
        """
        ...

    def split_old_comment(self) -> Any:
        """split the post part of a comment, and return it
        as comment to be added. Delete second part if [None, None]
         abc:  # this goes to sequence
           # this goes to first element
           - first element
        """
        ...

    def move_new_comment(self, target: Any, empty: bool = ...) -> Any:
        """move a comment from this token to target (normally next token)
        used to combine e.g. comments before a BlockEntryToken to the
        ScalarToken that follows it
        empty is a special for empty values -> comment after key
        """
        ...

class DirectiveToken(Token):
    __slots__ = ...
    id = ...
    def __init__(
        self, name: Any, value: Any, start_mark: Any, end_mark: Any
    ) -> None: ...

class DocumentStartToken(Token):
    __slots__ = ...
    id = ...

class DocumentEndToken(Token):
    __slots__ = ...
    id = ...

class StreamStartToken(Token):
    __slots__ = ...
    id = ...
    def __init__(
        self, start_mark: Any = ..., end_mark: Any = ..., encoding: Any = ...
    ) -> None: ...

class StreamEndToken(Token):
    __slots__ = ...
    id = ...

class BlockSequenceStartToken(Token):
    __slots__ = ...
    id = ...

class BlockMappingStartToken(Token):
    __slots__ = ...
    id = ...

class BlockEndToken(Token):
    __slots__ = ...
    id = ...

class FlowSequenceStartToken(Token):
    __slots__ = ...
    id = ...

class FlowMappingStartToken(Token):
    __slots__ = ...
    id = ...

class FlowSequenceEndToken(Token):
    __slots__ = ...
    id = ...

class FlowMappingEndToken(Token):
    __slots__ = ...
    id = ...

class KeyToken(Token):
    __slots__ = ...
    id = ...

class ValueToken(Token):
    __slots__ = ...
    id = ...

class BlockEntryToken(Token):
    __slots__ = ...
    id = ...

class FlowEntryToken(Token):
    __slots__ = ...
    id = ...

class AliasToken(Token):
    __slots__ = ...
    id = ...
    def __init__(self, value: Any, start_mark: Any, end_mark: Any) -> None: ...

class AnchorToken(Token):
    __slots__ = ...
    id = ...
    def __init__(self, value: Any, start_mark: Any, end_mark: Any) -> None: ...

class TagToken(Token):
    __slots__ = ...
    id = ...
    def __init__(self, value: Any, start_mark: Any, end_mark: Any) -> None: ...

class ScalarToken(Token):
    __slots__ = ...
    id = ...
    def __init__(
        self, value: Any, plain: Any, start_mark: Any, end_mark: Any, style: Any = ...
    ) -> None: ...

class CommentToken(Token):
    __slots__ = ...
    id = ...
    def __init__(
        self, value: Any, start_mark: Any = ..., end_mark: Any = ..., column: Any = ...
    ) -> None: ...
    @property
    def value(self) -> str: ...
    @value.setter
    def value(self, val: Any) -> None: ...
    def reset(self) -> None: ...
    def __repr__(self) -> Any: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...



"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.error import YAMLError
from ruamel.yaml.events import *

if False: ...
__all__ = ["Emitter", "EmitterError"]

class EmitterError(YAMLError): ...

class ScalarAnalysis:
    def __init__(
        self,
        scalar: Any,
        empty: Any,
        multiline: Any,
        allow_flow_plain: bool,
        allow_block_plain: bool,
        allow_single_quoted: bool,
        allow_double_quoted: bool,
        allow_block: bool,
    ) -> None: ...
    def __repr__(self) -> str: ...

class Indents:
    def __init__(self) -> None: ...
    def append(self, val: Any, seq: Any) -> None: ...
    def pop(self) -> Any: ...
    def seq_seq(self) -> bool: ...
    def last_seq(self) -> bool: ...
    def seq_flow_align(
        self, seq_indent: int, column: int, pre_comment: Optional[bool] = ...
    ) -> int: ...
    def __len__(self) -> int: ...

class Emitter:
    DEFAULT_TAG_PREFIXES = ...
    MAX_SIMPLE_KEY_LENGTH = ...
    flow_seq_start = ...
    flow_seq_end = ...
    flow_seq_separator = ...
    flow_map_start = ...
    flow_map_end = ...
    flow_map_separator = ...
    def __init__(
        self,
        stream: StreamType,
        canonical: Any = ...,
        indent: Optional[int] = ...,
        width: Optional[int] = ...,
        allow_unicode: Optional[bool] = ...,
        line_break: Any = ...,
        block_seq_indent: Optional[int] = ...,
        top_level_colon_align: Optional[bool] = ...,
        prefix_colon: Any = ...,
        brace_single_entry_mapping_in_flow_sequence: Optional[bool] = ...,
        dumper: Any = ...,
    ) -> None: ...
    @property
    def stream(self) -> Any: ...
    @stream.setter
    def stream(self, val: Any) -> None: ...
    @property
    def serializer(self) -> Any: ...
    @property
    def flow_level(self) -> int: ...
    def dispose(self) -> None: ...
    def emit(self, event: Any) -> None: ...
    def need_more_events(self) -> bool: ...
    def need_events(self, count: int) -> bool: ...
    def increase_indent(
        self, flow: bool = ..., sequence: Optional[bool] = ..., indentless: bool = ...
    ) -> None: ...
    def expect_stream_start(self) -> None: ...
    def expect_nothing(self) -> None: ...
    def expect_first_document_start(self) -> Any: ...
    def expect_document_start(self, first: bool = ...) -> None: ...
    def expect_document_end(self) -> None: ...
    def expect_document_root(self) -> None: ...
    def expect_node(
        self,
        root: bool = ...,
        sequence: bool = ...,
        mapping: bool = ...,
        simple_key: bool = ...,
    ) -> None: ...
    def expect_alias(self) -> None: ...
    def expect_scalar(self) -> None: ...
    def expect_flow_sequence(self, force_flow_indent: Optional[bool] = ...) -> None: ...
    def expect_first_flow_sequence_item(self) -> None: ...
    def expect_flow_sequence_item(self) -> None: ...
    def expect_flow_mapping(
        self, single: Optional[bool] = ..., force_flow_indent: Optional[bool] = ...
    ) -> None: ...
    def expect_first_flow_mapping_key(self) -> None: ...
    def expect_flow_mapping_key(self) -> None: ...
    def expect_flow_mapping_simple_value(self) -> None: ...
    def expect_flow_mapping_value(self) -> None: ...
    def expect_block_sequence(self) -> None: ...
    def expect_first_block_sequence_item(self) -> Any: ...
    def expect_block_sequence_item(self, first: bool = ...) -> None: ...
    def expect_block_mapping(self) -> None: ...
    def expect_first_block_mapping_key(self) -> None: ...
    def expect_block_mapping_key(self, first: Any = ...) -> None: ...
    def expect_block_mapping_simple_value(self) -> None: ...
    def expect_block_mapping_value(self) -> None: ...
    def check_empty_sequence(self) -> bool: ...
    def check_empty_mapping(self) -> bool: ...
    def check_empty_document(self) -> bool: ...
    def check_simple_key(self) -> bool: ...
    def process_anchor(self, indicator: Any) -> bool: ...
    def process_tag(self) -> None: ...
    def choose_scalar_style(self) -> Any: ...
    def process_scalar(self) -> None: ...
    def prepare_version(self, version: Any) -> Any: ...
    def prepare_tag_handle(self, handle: Any) -> Any: ...
    def prepare_tag_prefix(self, prefix: Any) -> Any: ...
    def prepare_tag(self, tag: Any) -> Any: ...
    def prepare_anchor(self, anchor: Any) -> Any: ...
    def analyze_scalar(self, scalar: Any) -> Any: ...
    def flush_stream(self) -> None: ...
    def write_stream_start(self) -> None: ...
    def write_stream_end(self) -> None: ...
    def write_indicator(
        self,
        indicator: Any,
        need_whitespace: Any,
        whitespace: bool = ...,
        indention: bool = ...,
    ) -> None: ...
    def write_indent(self) -> None: ...
    def write_line_break(self, data: Any = ...) -> None: ...
    def write_version_directive(self, version_text: Any) -> None: ...
    def write_tag_directive(self, handle_text: Any, prefix_text: Any) -> None: ...
    def write_single_quoted(self, text: Any, split: Any = ...) -> None: ...

    ESCAPE_REPLACEMENTS = ...
    def write_double_quoted(self, text: Any, split: Any = ...) -> None: ...
    def determine_block_hints(self, text: Any) -> Any: ...
    def write_folded(self, text: Any, comment: Any) -> None: ...
    def write_literal(self, text: Any, comment: Any = ...) -> None: ...
    def write_plain(self, text: Any, split: Any = ...) -> None: ...
    def write_comment(self, comment: Any, pre: bool = ...) -> None: ...
    def write_pre_comment(self, event: Any) -> bool: ...
    def write_post_comment(self, event: Any) -> bool: ...

class RoundTripEmitter(Emitter):
    def prepare_tag(self, ctag: Any) -> Any: ...



from __future__ import annotations

from typing import Any

import pytest
from hypothesis import given
from hypothesis import strategies as st
from pytest_mock import MockerFixture

from snakebids.plugins.version import Version, impm
from tests.helpers import allow_function_scoped


@given(version=st.text())
def test_explicit_version_used_directly(version: str):
    plug = Version(version=version)
    config: dict[Any, Any] = {}
    plug.initialize_config(config)
    assert config == {"plugins.version.version": version}


@given(version=st.text(), distribution=st.text())
@allow_function_scoped
def test_distribution_is_retrieved(
    version: str, distribution: str, mocker: MockerFixture
):
    mocker.stopall()
    mock = mocker.patch.object(impm, "version", return_value=version)
    plug = Version(distribution=distribution)
    config: dict[Any, Any] = {}
    plug.initialize_config(config)
    assert config == {"plugins.version.version": version}
    mock.assert_called_once_with(distribution)


@given(distribution=st.text())
@allow_function_scoped
def test_uninstalled_distribution_results_in_unknown(
    distribution: str, mocker: MockerFixture
):
    mocker.stopall()
    mock = mocker.patch.object(impm, "version", side_effect=impm.PackageNotFoundError)
    plug = Version(distribution=distribution)
    config: dict[Any, Any] = {}
    plug.initialize_config(config)
    assert plug.get(config, "version") == "unknown"
    mock.assert_called_once_with(distribution)


@given(version=st.text(), distribution=st.text())
def test_version_and_distribution_cannot_both_be_provided(
    version: str, distribution: str
):
    with pytest.raises(ValueError, match="version and distribution may not both"):
        Version(version=version, distribution=distribution)


def test_one_of_version_and_distribution_must_be_provided():
    with pytest.raises(ValueError, match="One of version or distribution"):
        Version()



"""
This type stub file was generated by pyright.
"""

from duecredit import due

"""

Stub file for a guaranteed safe import of duecredit constructs:  if duecredit
is not available.

To use it, place it into your project codebase to be imported, e.g. copy as

    cp stub.py /path/tomodule/module/due.py

Note that it might be better to avoid naming it duecredit.py to avoid shadowing
installed duecredit.

Then use in your code as

    from .due import due, Doi, BibTeX, Text

See  https://github.com/duecredit/duecredit/blob/master/README.md for examples.

Origin:     Originally a part of the duecredit
Copyright:  2015-2019  DueCredit developers
License:    BSD-2
"""
__version__ = ...

class InactiveDueCreditCollector:
    """Just a stub at the Collector which would not do anything"""

    def dcite(self, *args, **kwargs):  # -> (func: Unknown) -> Unknown:
        """If I could cite I would"""
        ...
    active = ...
    load = ...
    def __repr__(self): ...

if "due" in locals() and nothasattr(due, "cite"): ...



"""
This type stub file was generated by pyright.
"""

from _ruamel_yaml import CEmitter, CParser
from ruamel.yaml.constructor import BaseConstructor, Constructor, SafeConstructor
from ruamel.yaml.representer import BaseRepresenter, Representer, SafeRepresenter
from ruamel.yaml.resolver import BaseResolver, Resolver

if False: ...
__all__ = [
    "CBaseLoader",
    "CSafeLoader",
    "CLoader",
    "CBaseDumper",
    "CSafeDumper",
    "CDumper",
]

class CBaseLoader(CParser, BaseConstructor, BaseResolver):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...

class CSafeLoader(CParser, SafeConstructor, Resolver):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...

class CLoader(CParser, Constructor, Resolver):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...

class CBaseDumper(CEmitter, BaseRepresenter, BaseResolver):
    def __init__(
        self: StreamType,
        stream: Any,
        default_style: Any = ...,
        default_flow_style: Any = ...,
        canonical: Optional[bool] = ...,
        indent: Optional[int] = ...,
        width: Optional[int] = ...,
        allow_unicode: Optional[bool] = ...,
        line_break: Any = ...,
        encoding: Any = ...,
        explicit_start: Optional[bool] = ...,
        explicit_end: Optional[bool] = ...,
        version: Any = ...,
        tags: Any = ...,
        block_seq_indent: Any = ...,
        top_level_colon_align: Any = ...,
        prefix_colon: Any = ...,
    ) -> None: ...

class CSafeDumper(CEmitter, SafeRepresenter, Resolver):
    def __init__(
        self: StreamType,
        stream: Any,
        default_style: Any = ...,
        default_flow_style: Any = ...,
        canonical: Optional[bool] = ...,
        indent: Optional[int] = ...,
        width: Optional[int] = ...,
        allow_unicode: Optional[bool] = ...,
        line_break: Any = ...,
        encoding: Any = ...,
        explicit_start: Optional[bool] = ...,
        explicit_end: Optional[bool] = ...,
        version: Any = ...,
        tags: Any = ...,
        block_seq_indent: Any = ...,
        top_level_colon_align: Any = ...,
        prefix_colon: Any = ...,
    ) -> None: ...

class CDumper(CEmitter, Representer, Resolver):
    def __init__(
        self: StreamType,
        stream: Any,
        default_style: Any = ...,
        default_flow_style: Any = ...,
        canonical: Optional[bool] = ...,
        indent: Optional[int] = ...,
        width: Optional[int] = ...,
        allow_unicode: Optional[bool] = ...,
        line_break: Any = ...,
        encoding: Any = ...,
        explicit_start: Optional[bool] = ...,
        explicit_end: Optional[bool] = ...,
        version: Any = ...,
        tags: Any = ...,
        block_seq_indent: Any = ...,
        top_level_colon_align: Any = ...,
        prefix_colon: Any = ...,
    ) -> None: ...



from __future__ import annotations

from collections.abc import Iterator
from typing import TypeAlias

import importlib_resources as impr
import more_itertools as itx
from typing_extensions import NotRequired, TypedDict

from snakebids.io.yaml import get_yaml_io
from snakebids.paths import resources

__all__ = ["BidsPathEntitySpec", "BidsPathSpec", "BidsPathSpecFile"]


class BidsPathEntitySpec(TypedDict):
    """Defines an entity in a bids path."""

    entity: str
    """Entity full name"""

    tag: NotRequired[str]
    """Short entity name, as appears in the path"""

    dir: NotRequired[bool]
    """If true, a directory with the entity-value pair is created"""


BidsPathSpec: TypeAlias = list[BidsPathEntitySpec]
"""List of :class:`BidsPathEntitySpec`, defining the order of entities in a bids path"""


class BidsPathSpecFile(TypedDict):
    """Defines the valid structure for a BidsSpec definition file."""

    version: str
    """Version of the spec in semver"""

    description: NotRequired[str]
    """Optional description to be used in the spec function docstring"""

    spec: BidsPathSpec
    """Definition of the spec"""


def load_spec(path: impr.abc.Traversable) -> BidsPathSpecFile:
    """Return the spec saved in the provided yaml file."""
    return get_yaml_io().load(path.read_text())


def get_specs() -> Iterator[BidsPathSpecFile]:
    """Yield all defined bids spec file objects."""
    for path in impr.files(resources).iterdir():
        if (
            path.is_file()
            and path.name.startswith("spec.")
            and path.name.endswith(".yaml")
        ):
            yield load_spec(path)


def get_spec_path(version: str) -> impr.abc.Traversable:
    """Return path corresponding to provided spec version attribute.

    Parameters
    ----------
    version
        version attribute formatted as vx_x_x where x_x_x is spec semver
    """
    dotted = version[1:].replace("_", ".")
    return impr.files(resources).joinpath(f"spec.{dotted}.yaml")


def find_entity(spec: BidsPathSpec, entity: str) -> BidsPathEntitySpec:
    """Return configuration for specified entity out of BidsPathSpec."""
    return itx.one(item for item in spec if item["entity"] == entity)



"""
This type stub file was generated by pyright.
"""

from typing import Any, Dict, Optional, Union

from ruamel.yaml.error import *
from ruamel.yaml.nodes import *

if False: ...
__all__ = [
    "BaseRepresenter",
    "SafeRepresenter",
    "Representer",
    "RepresenterError",
    "RoundTripRepresenter",
]

class RepresenterError(YAMLError): ...

class BaseRepresenter:
    yaml_representers: Dict[Any, Any] = ...
    yaml_multi_representers: Dict[Any, Any] = ...
    def __init__(
        self: Any,
        default_style: Any = ...,
        default_flow_style: Any = ...,
        dumper: Any = ...,
    ) -> None: ...
    @property
    def serializer(self) -> Any: ...
    def represent(self, data: Any) -> None: ...
    def represent_data(self, data: Any) -> Any: ...
    def represent_key(self, data: Any) -> Any:
        """
        David Fraser: Extract a method to represent keys in mappings, so that
        a subclass can choose not to quote them (for example)
        used in represent_mapping
        https://bitbucket.org/davidfraser/pyyaml/commits/d81df6eb95f20cac4a79eed95ae553b5c6f77b8c
        """
        ...

    @classmethod
    def add_representer(cls, data_type: Any, representer: Any) -> None: ...
    @classmethod
    def add_multi_representer(cls, data_type: Any, representer: Any) -> None: ...
    def represent_scalar(
        self, tag: Any, value: Any, style: Any = ..., anchor: Any = ...
    ) -> ScalarNode: ...
    def represent_sequence(
        self, tag: Any, sequence: Any, flow_style: Any = ...
    ) -> SequenceNode: ...
    def represent_omap(
        self, tag: Any, omap: Any, flow_style: Any = ...
    ) -> SequenceNode: ...
    def represent_mapping(
        self, tag: Any, mapping: Any, flow_style: Any = ...
    ) -> MappingNode: ...
    def ignore_aliases(self, data: Any) -> bool: ...

class SafeRepresenter(BaseRepresenter):
    def ignore_aliases(self, data: Any) -> bool: ...
    def represent_none(self, data: Any) -> ScalarNode: ...
    def represent_str(self, data: Any) -> Any: ...
    def represent_binary(self, data: Any) -> ScalarNode: ...
    def represent_bool(self, data: Any, anchor: Optional[Any] = ...) -> ScalarNode: ...
    def represent_int(self, data: Any) -> ScalarNode: ...

    inf_value = ...
    def represent_float(self, data: Any) -> ScalarNode: ...
    def represent_list(self, data: Any) -> SequenceNode: ...
    def represent_dict(self, data: Any) -> MappingNode: ...
    def represent_ordereddict(self, data: Any) -> SequenceNode: ...
    def represent_set(self, data: Any) -> MappingNode: ...
    def represent_date(self, data: Any) -> ScalarNode: ...
    def represent_datetime(self, data: Any) -> ScalarNode: ...
    def represent_yaml_object(
        self, tag: Any, data: Any, cls: Any, flow_style: Any = ...
    ) -> MappingNode: ...
    def represent_undefined(self, data: Any) -> None: ...

class Representer(SafeRepresenter):
    def represent_complex(self, data: Any) -> Any: ...
    def represent_tuple(self, data: Any) -> SequenceNode: ...
    def represent_name(self, data: Any) -> ScalarNode: ...
    def represent_module(self, data: Any) -> ScalarNode: ...
    def represent_object(self, data: Any) -> Union[SequenceNode, MappingNode]: ...

class RoundTripRepresenter(SafeRepresenter):
    def __init__(
        self, default_style: Any = ..., default_flow_style: Any = ..., dumper: Any = ...
    ) -> None: ...
    def ignore_aliases(self, data: Any) -> bool: ...
    def represent_none(self, data: Any) -> ScalarNode: ...
    def represent_literal_scalarstring(self, data: Any) -> ScalarNode: ...

    represent_preserved_scalarstring = ...
    def represent_folded_scalarstring(self, data: Any) -> ScalarNode: ...
    def represent_single_quoted_scalarstring(self, data: Any) -> ScalarNode: ...
    def represent_double_quoted_scalarstring(self, data: Any) -> ScalarNode: ...
    def represent_plain_scalarstring(self, data: Any) -> ScalarNode: ...
    def insert_underscore(
        self, prefix: Any, s: Any, underscore: Any, anchor: Any = ...
    ) -> ScalarNode: ...
    def represent_scalar_int(self, data: Any) -> ScalarNode: ...
    def represent_binary_int(self, data: Any) -> ScalarNode: ...
    def represent_octal_int(self, data: Any) -> ScalarNode: ...
    def represent_hex_int(self, data: Any) -> ScalarNode: ...
    def represent_hex_caps_int(self, data: Any) -> ScalarNode: ...
    def represent_scalar_float(self, data: Any) -> ScalarNode:
        """This is way more complicated"""
        ...

    def represent_sequence(
        self, tag: Any, sequence: Any, flow_style: Any = ...
    ) -> SequenceNode: ...
    def merge_comments(self, node: Any, comments: Any) -> Any: ...
    def represent_key(self, data: Any) -> Any: ...
    def represent_mapping(
        self, tag: Any, mapping: Any, flow_style: Any = ...
    ) -> MappingNode: ...
    def represent_omap(
        self, tag: Any, omap: Any, flow_style: Any = ...
    ) -> SequenceNode: ...
    def represent_set(self, setting: Any) -> MappingNode: ...
    def represent_dict(self, data: Any) -> MappingNode:
        """Write out tag if saved on loading"""
        ...

    def represent_list(self, data: Any) -> SequenceNode: ...
    def represent_datetime(self, data: Any) -> ScalarNode: ...
    def represent_tagged_scalar(self, data: Any) -> ScalarNode: ...
    def represent_scalar_bool(self, data: Any) -> ScalarNode: ...
    def represent_yaml_object(
        self, tag: Any, data: Any, cls: Any, flow_style: Optional[Any] = ...
    ) -> MappingNode: ...



"""Helper functions and classes for tests"""

from __future__ import annotations

import functools as ft
import itertools as it
import re
import subprocess as sp
from collections.abc import Callable, Iterable, Mapping, Sequence
from datetime import timedelta
from pathlib import Path
from typing import (
    Any,
    Protocol,
    TypeVar,
)

import more_itertools as itx
import pytest
from hypothesis import HealthCheck, example, settings
from typing_extensions import ParamSpec

from snakebids import bids_factory
from snakebids.core.datasets import BidsDataset
from snakebids.core.input_generation import generate_inputs
from snakebids.paths import specs
from snakebids.types import InputsConfig, ZipList, ZipListLike
from snakebids.utils.containers import MultiSelectDict
from snakebids.utils.utils import BidsEntity

_T = TypeVar("_T")
_T_contra = TypeVar("_T_contra", contravariant=True)
_P = ParamSpec("_P")


def get_zip_list(
    entities: Iterable[BidsEntity | str], combinations: Iterable[tuple[str, ...]]
) -> ZipList:
    """Return a zip list from iterables of entities and value combinations

    Parameters
    ----------
    entities : Iterable[str]
        Iterable of entities
    combinations : Iterable[tuple[str, ...]]
        Iterable of combinations (e.g. produced using itertools.product). Each tuple
        must be the same length as entities

    Returns
    -------
    dict[str, list[str]]
        zip_list representation of entity-value combinations
    """

    strict = True
    lists: Iterable[Sequence[str]] = list(zip(*combinations, strict=True))
    if not lists:
        strict = False
        lists = itx.repeatfunc(list[str])
    return MultiSelectDict(
        {
            BidsEntity(str(entity)).wildcard: list(combs)
            for entity, combs in zip(entities, lists, strict=strict)
        }
    )


def setify(dic: dict[Any, list[Any]]) -> dict[Any, set[Any]]:
    """Convert a dict of *->lists into a dict of *->sets

    Parameters
    ----------
    dic : dict[Any, list[Any]]
        Dict of lists to convert

    Returns
    -------
    dict[Any, Set[Any]]
        Dict of sets
    """
    return {key: set(val) for key, val in dic.items()}


def get_bids_path(entities: Iterable[str | BidsEntity], **extras: str) -> str:
    """Get consistently ordered bids path for a group of entities

    Parameters
    ----------
    entities : Iterable[str]
        Entities to convert into a bids path
    extras
        Extra entity-value pairs to include in the path

    Returns
    -------
    str
        bids path
    """

    def get_tag(entity: BidsEntity) -> tuple[str, str]:
        return entity.wildcard, f"{{{entity.wildcard}}}"

    bids = bids_factory(specs.latest())
    return bids(
        **dict(get_tag(BidsEntity(entity)) for entity in sorted(entities)),
        **{BidsEntity(entity).wildcard: value for entity, value in extras.items()},
    )


class BidsListCompare(dict[str, dict[str, list[str]]]):
    """Dict override specifically for comparing input_lists

    When comparing, all lists are converted into sets so that order doesn't matter for
    the comparison
    """

    def __eq__(self, other: object) -> bool:
        # Getting pyright to like this is more work than it's worth
        if not isinstance(other, dict):
            return False
        for name, lists in other.items():  # type: ignore
            if name not in self:
                return False
            for key, val in lists.items():  # type: ignore
                if (
                    not isinstance(val, list)
                    or key not in self[name]  # type: ignore
                    or set(self[name][key]) != set(val)  # type: ignore
                ):
                    return False
        return True


def debug(**overrides: Any):
    """Disable a hypothesis decorated test with specific parameter examples

    Should be used as a decorator, placed *before* @hypothesis.given(). Adds the "debug"
    mark to the test, which can be detected by other constructs (e.g. to disable fakefs
    when hypothesis is not being used)
    """

    def inner(func: Callable[_P, _T]) -> Callable[_P, _T]:
        if not hasattr(func, "hypothesis"):
            msg = f"{func} is not decorated with hypothesis.given"
            raise TypeError(msg)

        test = func.hypothesis.inner_test  # type: ignore

        @ft.wraps(func)
        def inner_test(*args: _P.args, **kwargs: _P.kwargs):
            return test(*args, **(kwargs | overrides))

        return inner_test

    return inner


def mock_data(*draws: Any):
    """Utility function for mocking the hypothesis data strategy

    Intended for combination with debug. Takes a list of values corresponding the draws
    taking from the data object. In the debug run, calls to data will return the given
    values in the order specified
    """

    # pylint: disable=missing-class-docstring, too-few-public-methods
    class MockData:
        _draws = iter(draws)

        # pylint: disable=unused-argument
        def draw(self, strategy: Any, label: Any = None):
            return next(self._draws)

    return MockData()


def create_dataset(root: str | Path, dataset: BidsDataset) -> None:
    """Create an empty BidsDataset on the filesystem

    Creates the directory structure and files represented by a BidsDataset. Files are
    touched: they will have no contents.
    """
    for component in dataset.values():
        for path in map(Path(root).joinpath, component.expand()):
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch()


def create_snakebids_config(dataset: BidsDataset) -> InputsConfig:
    """Generate a basic snakebids config dict from a dataset"""
    all_entities = set(
        it.chain.from_iterable(comp.entities for comp in dataset.values())
    )
    return {
        comp.name: {
            "filters": {
                BidsEntity.from_tag(entity).entity: list(comp.entities[entity])
                if entity in comp.entities
                else False
                for entity in all_entities
            },
            "wildcards": [
                BidsEntity.from_tag(entity).entity for entity in comp.entities
            ],
        }
        for comp in dataset.values()
    }


def reindex_dataset(
    root: str,
    dataset: BidsDataset,
    use_custom_paths: bool = False,
    participant_label: str | Sequence[str] | None = None,
    exclude_participant_label: str | Sequence[str] | None = None,
) -> BidsDataset:
    """Create BidsDataset on the filesystem and reindex

    Paths within the dataset must be absolute
    """
    create_dataset(Path("/"), dataset)
    config = create_snakebids_config(dataset)
    if participant_label is not None or exclude_participant_label is not None:
        for comp in config.values():
            if "subject" in comp.get("filters", {}):
                del comp["filters"]["subject"]  # type: ignore

    if use_custom_paths:
        for comp in config:
            config[comp]["custom_path"] = dataset[comp].path
    return generate_inputs(
        root,
        config,
        participant_label=participant_label,
        exclude_participant_label=exclude_participant_label,
    )


def allow_function_scoped(func: _T, /) -> _T:
    """Allow function_scoped fixtures in hypothesis tests

    This is primarily useful for using tmpdirs, hence, the name
    """
    return settings(
        suppress_health_check=[
            HealthCheck.function_scoped_fixture,
        ],
    )(func)


def deadline(time: int | float | timedelta | None) -> Callable[[_T], _T]:
    """Change hypothesis deadline

    Numbers refer to time in milliseconds. Set to None to disable entirely
    """

    def inner(func: _T, /) -> _T:
        return settings(deadline=time)(func)

    return inner


def expand_zip_list(
    zip_list: ZipListLike, new_values: Mapping[str, Sequence[str]]
) -> ZipListLike:
    """Expand a zip list with new values via product

    Each zip list column will be combined with all possible combinations of the
    entity-value pairs in ``new_values``. It is thus equivalent to calling
    :meth:`BidsComponent.expand() <snakebids.BidsComponent.expand>` with ``new_values``
    as extra args.
    """
    zip_cols = list(zip(*zip_list.values(), strict=True))
    new_cols = list(
        map(
            it.chain.from_iterable,
            it.product(zip_cols, it.product(*new_values.values())),
        )
    )
    z = zip(*new_cols, strict=True)
    return dict(zip(it.chain(zip_list.keys(), new_values.keys()), z, strict=True))


def needs_docker(container: str):
    def decorator(func: Callable[_P, _T]) -> Callable[_P, _T]:
        @pytest.mark.docker
        @ft.wraps(func)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs):
            try:
                sp.run(["docker"], check=True, capture_output=True)
            except (sp.CalledProcessError, FileNotFoundError):
                pytest.fail(
                    "docker is not available on this machine. To skip docker tests, "
                    "use '-m \"not docker\"'"
                )
            try:
                sp.run(
                    ["docker", "image", "inspect", container],
                    capture_output=True,
                    check=True,
                )
            except sp.CalledProcessError as err:
                if not (
                    match := re.match(r"snakebids/([\w\-]+)\:[a-zA-Z0-9\-]+", container)
                ):
                    pytest.fail(
                        f"Unrecognized docker image: {container}. Should be "
                        "'snakebids/{container_id}:{version}"
                    )
                container_id = match[1]
                pytest.fail(
                    f"{container} is not built on this machine. To build container, "
                    f"run:\n\n\t`poetry run poe build-container {container_id}`\n\nTo "
                    f"skip docker tests, use '-m \"not docker\"'.\n\n(error:) {err}"
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def identity(obj: _T) -> _T:
    return obj


def example_if(condition: bool, *args: Any, **kwargs: Any):
    def inner(func: Callable[_P, _T]) -> Callable[_P, _T]:
        return example(*args, **kwargs)(func)

    if condition:
        return inner

    return identity


class Benchmark(Protocol):
    def __call__(
        self, func: Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs
    ) -> _T: ...


"""Comparison Dunders copied from typeshed"""


class SupportsDunderLT(Protocol[_T_contra]):
    def __lt__(self, __other: _T_contra) -> bool: ...


def is_strictly_increasing(items: Iterable[SupportsDunderLT[Any]]) -> bool:
    # itx.pairwise properly aliases it.pairwise on py310+
    return all(i < j for i, j in itx.pairwise(items))



from __future__ import annotations

import argparse
import importlib
import re
from pathlib import Path
from typing import Any, cast

import attrs

from snakebids import bidsapp
from snakebids.exceptions import ConfigError
from snakebids.io.yaml import get_yaml_io


def _make_underscore_dash_aliases(name: str) -> set[str]:
    """Generate --dash-arg aliases for --dash_args and vice versa.

    If no dashes or underscores are in the argument name, a tuple containing just the
    original arg name will be returned

    Parameters
    ----------
    name : str
        Argument to generate conversion for

    Returns
    -------
    set of strings
        Converted args
    """
    if match := re.match(r"^--(.+)$", name):
        name_part = match.group(1)
        return {
            name,
            re.sub(r"\_", "-", name),
            f"--{re.sub(r'-', '_', name_part)}",
        }
    return {name}


def _find_type(name: str, *, yamlsafe: bool = True) -> type[Any]:
    if name == "Path":
        return Path
    *module_name, obj_name = name.split(".") if "." in name else ("builtins", name)
    try:
        type_ = getattr(importlib.import_module(".".join(module_name)), obj_name)
    except (ImportError, AttributeError) as err:
        msg = f"{name} could not be resolved"
        raise ConfigError(msg) from err
    if not callable(type_):
        msg = f"{name} cannot be used as a type"
        raise ConfigError(msg)
    if yamlsafe and type_ not in get_yaml_io().representer.yaml_representers:
        msg = f"{name} cannot be serialized into yaml"
        raise ConfigError(msg)
    return cast("type[Any]", type_)


@attrs.define
class CliConfig:
    """Configure CLI arguments directly in config.

    Arguments are provided in config in a dictionary stored under ``cli_key``. Each
    entry maps the name of the argument to a set of valid arguments for
    :meth:`~argparse.ArgumentParser.add_argument()`.

    This plugin will attempt to be the first to add arguments, and thus can be used
    to override arguments from other compatible plugins, such as
    :class:`~snakebids.plugins.bidsargs.BidsArgs`

    Parameters
    ----------
    cli_key
        Key of dict containing arguments

    Example
    -------

    .. code-block:: yaml

        parse_args:
            --tunable-parameter:
                help: |
                    A parameter important to the analysis that you can be set from the
                    commandline. If not set, a sensible default will be used
                default: 5
                type: float
            --alternate-mode:
                help: |
                    A flag activating a secondary feature of the workflow
                action: store_true
            --derivatives:
                help: |
                    An alternate help message for --derivatives, which will override
                    the help message from argument given by ``BidsArgs``. Note that we
                    must again specify the ``nargs`` and ``type`` for the param
                type: Path
                nargs: "*"
    """

    cli_key: str = "parse_args"

    def __eq__(self, other: Any):
        return isinstance(other, self.__class__)

    @bidsapp.hookimpl(tryfirst=True)
    def add_cli_arguments(
        self, parser: argparse.ArgumentParser, config: dict[str, Any]
    ):
        """Add arguments from config."""
        if (parse_args := config.get(self.cli_key)) is None:
            return

        # update the parser with config options
        for name, arg in parse_args.items():
            # Convert type annotations from strings to class types
            # We first check that the type annotation is, in fact,
            # a str to allow the edge case where it's already
            # been converted
            if "type" in arg:
                arg_dict = {**arg, "type": _find_type(str(arg["type"]))}
            else:
                arg_dict = arg
            parser.add_argument(
                *_make_underscore_dash_aliases(name),
                **arg_dict,  # type: ignore
            )



# This stub file is automatically generated
# It can be updated using::
#
#      poetry run poe update_bids

from ._utils import BidsPathSpec

LATEST: str

def v0_0_0(subject_dir: bool = True, session_dir: bool = True) -> BidsPathSpec:
    """Get the v0.0.0 BidsPathSpec.

    This spec alone equips :func:`~snakebids.bids` with 2 extra arguments:
    ``include_subject_dir`` and ``include_session_dir``. These default to ``True``, but
    if set ``False``, remove the subject and session dirs respectively from the output
    path. For future specs, this behaviour should be achieved by modifying the spec and
    generating a new :func:`~snakebids.bids` function

    Formatted as::

        sub-{subject}/ses-{session}/{datatype}/{prefix}_sub-{subject}_ses-{session}_
        task-{task}_acq-{acq}_ce-{ce}_rec-{rec}_dir-{dir}_run-{run}_mod-{mod}_
        echo-{echo}_hemi-{hemi}_space-{space}_res-{res}_den-{den}_label-{label}_
        desc-{desc}_..._{suffix}{extension}


    Parameters
    ----------
    subject_dir
        If False, downstream path generator will not include the subject dir
        `sub-{subject}/*`
    session_dir : bool, optional
        If False, downstream path generator will not include the session dir
        `*/ses-{session}/*`
    """

def v0_11_0(subject_dir: bool = True, session_dir: bool = True) -> BidsPathSpec:
    """Spec corresponding to `BIDS v1.9.0`_.

    Significantly expanded from the v0.0.0 spec, now including long names for every
    relevant entity. In addition to the official spec, it includes `from` and `to`
    entities intended for transformations. Unknown entities are placed just before desc,
    so that the description entity is always last.

    .. _BIDS v1.9.0: https://bids-specification.readthedocs.io/en/v1.9.0/

    Formatted as::

        sub-{subject}/ses-{session}/{datatype}/{prefix}_sub-{subject}_ses-{session}_
        sample-{sample}_task-{task}_tracksys-{tracksys}_acq-{acquisition}_
        ce-{ceagent}_stain-{staining}_trc-{tracer}_rec-{reconstruction}_
        dir-{direction}_run-{run}_mod-{modality}_echo-{echo}_flip-{flip}_
        inv-{inversion}_mt-{mt}_proc-{processed}_part-{part}_space-{space}_
        atlas-{atlas}_seg-{segmentation}_hemi-{hemisphere}_res-{resolution}_
        den-{density}_roi-{roi}_from-{from}_to-{to}_split-{split}_
        recording-{recording}_chunk-{chunk}_model-{model}_subset-{subset}_
        label-{label}_..._desc-{description}_{suffix}{extension}


    Parameters
    ----------
    subject_dir
        If False, downstream path generator will not include the subject dir
        `sub-{subject}/*`
    session_dir : bool, optional
        If False, downstream path generator will not include the session dir
        `*/ses-{session}/*`
    """

def v0_15_0(subject_dir: bool = True, session_dir: bool = True) -> BidsPathSpec:
    """Spec corresponding to `BIDS v1.10.1`_.

    In addition to the official spec, it includes `from` and `to` entities intended for
    transformations. Unknown entities are placed just before desc, so that the
    description entity is always last.

    .. _BIDS v1.10.1: https://bids-specification.readthedocs.io/en/v1.10.1/

    Formatted as::

        sub-{subject}/ses-{session}/{datatype}/{prefix}_sub-{subject}_ses-{session}_
        sample-{sample}_task-{task}_tracksys-{tracksys}_acq-{acquisition}_
        nuc-{nucleus}_voi-{volume}_ce-{ceagent}_stain-{staining}_trc-{tracer}_
        rec-{reconstruction}_dir-{direction}_run-{run}_mod-{modality}_echo-{echo}_
        flip-{flip}_inv-{inversion}_mt-{mt}_proc-{processed}_part-{part}_
        space-{space}_atlas-{atlas}_seg-{segmentation}_hemi-{hemisphere}_
        res-{resolution}_den-{density}_roi-{roi}_from-{from}_to-{to}_split-{split}_
        recording-{recording}_chunk-{chunk}_model-{model}_subset-{subset}_
        label-{label}_..._desc-{description}_{suffix}{extension}


    Parameters
    ----------
    subject_dir
        If False, downstream path generator will not include the subject dir
        `sub-{subject}/*`
    session_dir : bool, optional
        If False, downstream path generator will not include the session dir
        `*/ses-{session}/*`
    """

def latest(subject_dir: bool = True, session_dir: bool = True) -> BidsPathSpec:
    """Spec corresponding to `BIDS v1.10.1`_.

    In addition to the official spec, it includes `from` and `to` entities intended for
    transformations. Unknown entities are placed just before desc, so that the
    description entity is always last.

    .. _BIDS v1.10.1: https://bids-specification.readthedocs.io/en/v1.10.1/

    Formatted as::

        sub-{subject}/ses-{session}/{datatype}/{prefix}_sub-{subject}_ses-{session}_
        sample-{sample}_task-{task}_tracksys-{tracksys}_acq-{acquisition}_
        nuc-{nucleus}_voi-{volume}_ce-{ceagent}_stain-{staining}_trc-{tracer}_
        rec-{reconstruction}_dir-{direction}_run-{run}_mod-{modality}_echo-{echo}_
        flip-{flip}_inv-{inversion}_mt-{mt}_proc-{processed}_part-{part}_
        space-{space}_atlas-{atlas}_seg-{segmentation}_hemi-{hemisphere}_
        res-{resolution}_den-{density}_roi-{roi}_from-{from}_to-{to}_split-{split}_
        recording-{recording}_chunk-{chunk}_model-{model}_subset-{subset}_
        label-{label}_..._desc-{description}_{suffix}{extension}


    Parameters
    ----------
    subject_dir
        If False, downstream path generator will not include the subject dir
        `sub-{subject}/*`
    session_dir : bool, optional
        If False, downstream path generator will not include the session dir
        `*/ses-{session}/*`
    """



from os import PathLike
from typing import Any, AnyStr, Literal, Mapping, TypedDict

from git import Sequence
from typing_extensions import TypeAlias, Unpack

AnyPath: TypeAlias = "AnyStr | PathLike[AnyStr]"

class CopierArgs(TypedDict, total=False):
    answers_file: AnyPath[str] | AnyPath[bytes]
    vcs_ref: str | None
    exclude: Sequence[str]
    use_prereleases: bool
    skip_if_exists: Sequence[str]
    cleanup_on_error: bool
    defaults: bool
    user_defaults: dict[str, Any]
    overwrite: bool
    pretend: bool
    quiet: bool
    conflict: Literal["inline", "rej"]
    context_lines: int
    unsafe: bool
    skip_answered: bool

def run_copy(
    src_path: str,
    dst_path: AnyPath[str] | AnyPath[bytes] = ...,
    data: Mapping[str, Any] | None = ...,
    **kwargs: Unpack[CopierArgs]
) -> None: ...



"""
This type stub file was generated by pyright.
"""

if False: ...
tag_attrib = ...

class Tag:
    """store original tag information for roundtripping"""

    attrib = ...
    def __init__(
        self, handle: Any = ..., suffix: Any = ..., handles: Any = ...
    ) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: Any) -> bool: ...
    def startswith(self, x: str) -> bool: ...
    @property
    def trval(self) -> Optional[str]: ...

    value = ...
    @property
    def uri_decoded_suffix(self) -> Optional[str]: ...
    def select_transform(self, val: bool) -> None:
        """
        val: False -> non-round-trip
             True -> round-trip
        """
        ...

    def check_handle(self) -> bool: ...



"""
This type stub file was generated by pyright.
"""

import collections.abc
import io
import os
from abc import abstractmethod
from collections import OrderedDict
from typing import (
    Any,
    Optional,
    Union,
)

from ruamel.yaml.docinfo import Version

try:
    from typing import SupportsIndex as SupportsIndex  # in order to reexport for mypy
except ImportError:
    SupportsIndex = int  # type: ignore

StreamType = Any
StreamTextType = StreamType
VersionType = Union[str, tuple[int, int], list[int], Version, None]
_DEFAULT_YAML_VERSION = ...

class ordereddict(OrderedDict):
    if not hasattr(OrderedDict, "insert"):
        def insert(self, pos: int, key: Any, value: Any) -> None: ...

StringIO = io.StringIO
BytesIO = io.BytesIO
builtins_module = ...

def with_metaclass(meta: Any, *bases: Any) -> Any:
    """Create a base class with a metaclass."""
    ...

DBG_TOKEN = ...
DBG_EVENT = ...
DBG_NODE = ...
_debug: Optional[int] = ...
if "RUAMELDEBUG" in os.environ:
    _debugx = ...
if bool(_debug):
    class ObjectCounter:
        def __init__(self) -> None: ...
        def __call__(self, k: Any) -> None: ...
        def dump(self) -> None: ...

    object_counter = ...

def dbg(val: Any = ...) -> Any: ...

class Nprint:
    def __init__(self, file_name: Any = ...) -> None: ...
    def __call__(self, *args: Any, **kw: Any) -> None: ...
    def set_max_print(self, i: int) -> None: ...
    def fp(self, mode: str = ...) -> Any: ...

nprint = ...
nprintf = ...

def check_namespace_char(ch: Any) -> bool: ...
def check_anchorname_char(ch: Any) -> bool: ...
def version_tnf(t1: Any, t2: Any = ...) -> Any:
    """
    return True if ruamel.yaml version_info < t1, None if t2 is specified and bigger else False
    """
    ...

class MutableSliceableSequence(collections.abc.MutableSequence):
    __slots__ = ...
    def __getitem__(self, index: Any) -> Any: ...
    def __setitem__(self, index: Any, value: Any) -> None: ...
    def __delitem__(self, index: Any) -> None: ...
    @abstractmethod
    def __getsingleitem__(self, index: Any) -> Any: ...
    @abstractmethod
    def __setsingleitem__(self, index: Any, value: Any) -> None: ...
    @abstractmethod
    def __delsingleitem__(self, index: Any) -> None: ...



"""
This type stub file was generated by pyright.
"""

""" Exceptions.

Exceptions relating to problems with BIDS itself, should carry BIDS in their
name.  All exceptions should subclass from PyBIDSError
"""

class PyBIDSError(Exception):
    """Base class.  Typically for mix-in."""

    ...

class ConfigError(ValueError, PyBIDSError):
    """Problems with config file."""

    ...

class NoMatchError(ValueError, PyBIDSError):
    """No match found where it is required."""

    ...

class BIDSEntityError(AttributeError, PyBIDSError):
    """An unknown entity."""

    ...

class TargetError(ValueError, PyBIDSError):
    """An unknown target."""

    ...

class BIDSValidationError(ValueError, PyBIDSError):
    """An invalid BIDS dataset."""

    ...

class BIDSDerivativesValidationError(BIDSValidationError):
    """An invalid BIDS derivative dataset."""

    ...

class BIDSConflictingValuesError(BIDSValidationError):
    """A value conflict (e.g. in filename and side-car .json)"""

    ...
